/****************************************************************************
  FERAL ROBOT - URBAN TAPESTRIES (http://urbantapestries.org/)
  Linux-based robot client for gumstix platform
  Author: Dima Diall
****************************************************************************/

#include "config.h"
#include "utrobot.h"
#include "logger.h"
#include "adchw.h"
#include "sensordef.h"
#include <gps.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>

// Output streams for logging normal data, errors and debugging info
#define DAT         stdout
#define DBG         stderr
#define ERR         stderr

// X = online?, S = fix status?, M = fix mode, P = position, W = watcher [see gpsd(8)]
#define GPS_INIT_STRING    "XW+"

static void update_position(struct gps_data_t *gpsdata, 
                            char *message, size_t len, int level)
{
    log_msg(DBG, LL_TRACE, "GPS update: stat=%u, mode=%u, latitude=%.2f, longitude=%.2f\nmessage='%s'\n",
            gpsdata->status, gpsdata->fix.mode,
            gpsdata->fix.latitude, gpsdata->fix.longitude,
            message);
    
     /*
     // update global variables
     gps_fix = (gpsdata->fix.mode >= MODE_2D);
     if(gps_fix) {
         gps_latitude = gpsdata->fix.latitude;
         gps_longitude = gpsdata->fix.longitude;
     }
     */
}

int send_packet(int sock_fd, const packet_t *packet, const struct sockaddr_in *dst_addr)
{
    static packet_t buffer;
    int num_bytes;
    
    // copy byte array fields into transmission buffer
    bcopy(packet->robot_id, buffer.robot_id, sizeof(buffer.robot_id));
    bcopy(packet->sensor_id, buffer.sensor_id, sizeof(buffer.sensor_id));
    
    // convert integer binary fields to network byte order
    buffer.status = htons(packet->status);
    buffer.timestamp = htonl(packet->timestamp);
    
    // apply hack to correct floats to network byte order on little endian systems
    buffer.sensor_val.l = htonl(packet->sensor_val.l);
    buffer.latitude.l = htonl(packet->latitude.l);
    buffer.longitude.l = htonl(packet->longitude.l);
    
    log_msg(DBG, LL_TRACE, "Packet:");
    log_hex(DBG, LL_TRACE, &buffer, sizeof(buffer));
    
    num_bytes = sendto(sock_fd, &buffer, sizeof(buffer), 0,
                       (struct sockaddr *) dst_addr, sizeof(struct sockaddr_in));    
    
    return (num_bytes == sizeof(buffer) ? 0 : -1);
}

// TODO: provide real operating system retrieval of MAC address
// Dummy function to return MAC address string (...)
int get_mac_address(unsigned long long addr_bin, char *mac_addr, int len)
{
    uint16_t hi = htons((addr_bin >> 32) & 0xFFFF);
    uint32_t lo = htonl(addr_bin & 0xFFFFFFFF);
    char *hi_p = (char *) &hi;
    char *lo_p = (char *) &lo;
    int i;
    
    if(mac_addr == NULL || len < LEN_MAC)
        return -1;
    
    mac_addr[0] = hi_p[0];
    mac_addr[1] = hi_p[1];
    mac_addr[2] = lo_p[0];
    mac_addr[3] = lo_p[1];
    mac_addr[4] = lo_p[2];
    mac_addr[5] = lo_p[3];
    
    return 0;
}

char *cut_server_port(char *server_port)
{
    char *colon = strchr(server_port, ':');
    
    if (colon != NULL) {
        // cut original string at ':' and return pointer to substring after colon
        *colon = '\0';
        return colon + 1;
    }
    return NULL;
}

#define GETOPT_STRING       "i:u:g:a:lvdVh"

void usage(char *progname) 
{
    fprintf(stderr, 
            "Usage: %s -i hwaddr -u addr[:port] [-g addr[:port]] [-a serial_port] [-ldvh]\n\n"
            "  -i hwaddr        Robot id as MAC address (6 bytes in hex)\n"
            "  -u addr[:port]   UT server (default port: %u)\n"
            "  -g addr[:port]   GPSD server (default: " GPSD_ADDR ":" GPSD_PORT ")\n"
            "  -a serial_port   Serial port for ADC access (default: " ADC_SERIAL_PORT ")\n"
            "  -l               Log sensor data to standard output (stdout)\n"
            "  -v               Print extra information (on stderr)\n"
            "  -d               Print debugging information (on stderr)\n"
            "  -V               Show version, then exit\n"
            "  -h               Show this help message, then exit\n\n"
            "More information on Feral Robots at http://socialtapestries.net/feralrobots/\n"
            , progname, UT_SERVER_PORT);
    
    exit(1);
}

int main(int argc, char **argv)
{
    int option, i = 0;
    int sock_fd, adc_fd;
    struct sockaddr_in dst_addr;
    struct hostent *hostinfo;
    struct gps_data_t *gpsdata;
    unsigned long long robot_id = 0x0;
    char *utsrv_addr = NULL;
    char *utsrv_port = NULL;
    unsigned utsrv_port_num = UT_SERVER_PORT;
    char *gpsd_addr = GPSD_ADDR;
    char *gpsd_port = GPSD_PORT;
    char *adc_serial = ADC_SERIAL_PORT;
    char ip_str[INET_ADDRSTRLEN];
    uint16_t adc_val;
    int sensor_proc;
    packet_t packet;

    // initialize logger with minimal output (errors and warnings)
    logger_level = LL_FATAL | LL_ERROR | LL_WARN;

    // process command line arguments
    while((option = getopt(argc, argv, GETOPT_STRING)) != -1) {
        switch(option) {
            case 'u':
                utsrv_addr = optarg;
                if(utsrv_port = cut_server_port(utsrv_addr)) {
                    if(strlen(utsrv_addr) < 1)
                        utsrv_addr = NULL;
                    utsrv_port_num = strtoul(utsrv_port, NULL, 10);
                }
                break;
                
            case 'g':
                gpsd_addr = optarg;
                if(gpsd_port = cut_server_port(gpsd_addr)) {
                    if(strlen(gpsd_addr) < 1 || strlen(gpsd_port) < 1)
                        gpsd_addr = gpsd_port = NULL;
                } else {
                    gpsd_port = GPSD_PORT;
                }
                break;
                
            case 'i': robot_id   = strtoull(optarg, NULL, 16);  break;
            case 'a': adc_serial = optarg;                      break;

            case 'l': logger_level |= LL_INFO;                  break;
            case 'v': logger_level |= LL_DEBUG;                 break;
            case 'd': logger_level |= LL_DEBUG | LL_TRACE;      break;
                
            case 'V':
                fprintf(stderr, PACKAGE " " VERSION "\n");
                exit(0);
                break;
                
            case 'h':
            default:
                usage(argv[0]);
                break;
        }
    }

    log_msg(DBG, LL_TRACE,
            "Finished processing command-line options:\n"
            "Logger level:     %02x\n"
            "Robot id:         %08llx\n"
            "UT server:        %s:%d\n"
            "GPSD server:      %s:%s\n"
            "ADC serial port:  %s\n\n",
            logger_level, robot_id, utsrv_addr, utsrv_port_num, gpsd_addr, gpsd_port, adc_serial);
    
    // verify if the parameters for IP addresses and port numbers are correct
    if(robot_id == 0) {
        log_msg(ERR, LL_FATAL, "Invalid robot id (specify MAC address in hex, eg: 11a0f89fbec3).\n");
        usage(argv[0]);
    }
    if(utsrv_addr == NULL || utsrv_port_num == 0) {
        log_msg(ERR, LL_FATAL, "Invalid IP address or port number specified for UT server.\n");
        usage(argv[0]);
    }
    if(gpsd_addr == NULL || gpsd_port == NULL) {
        log_msg(ERR, LL_FATAL, "Invalid IP address or port number specified for GPSD server.\n");
        usage(argv[0]);
    }
    
    // open serial port to interface ADC device
    log_msg(DBG, LL_DEBUG, "Opening serial port to access ADC on %s...\n", adc_serial);
    adc_fd = open(adc_serial, O_RDWR);
    if(adc_fd == -1) {
        log_err(ERR, LL_FATAL, "Unable to open serial port to access ADC (%s).\n", adc_serial);
        exit(2);
    }
    
    // initialize connection to GPS daemon
    log_msg(DBG, LL_DEBUG, "Connecting to GPSD server at %s:%s...\n", gpsd_addr, gpsd_port);
    gpsdata = gps_open(gpsd_addr, gpsd_port);
    if(!gpsdata) {
        log_err(ERR, LL_FATAL, "Unable to connect to GPSD server (%s:%s).\n", gpsd_addr, gpsd_port);
        exit(2);
    }
    // register hook to position updates handler and init watcher mode
    gps_set_raw_hook(gpsdata, update_position);
    gps_query(gpsdata, GPS_INIT_STRING "\n");
    gps_poll(gpsdata);

    // create a datagram-based socket (rx/tx)
    log_msg(DBG, LL_DEBUG, "Creating UDP socket and resolving host address '%s'...\n", utsrv_addr);
    if((sock_fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
        log_err(ERR, LL_FATAL, "Unable to create local UDP socket.\n");
        exit(2);
    }

    // TODO: add support for IPv6 addresses (future?)
    
    // set-up IP address & UDP port for sending messages
    bzero(&dst_addr, sizeof(dst_addr));
    dst_addr.sin_family = AF_INET;
    dst_addr.sin_port = htons(utsrv_port_num);

    // resolve hostname to IP address
    if((hostinfo = gethostbyname(utsrv_addr)) == NULL) {
        // if fails to resolve, attempt to convert IP address directly
        if(inet_aton(utsrv_addr, &dst_addr.sin_addr) == 0) {
            log_msg(ERR, LL_FATAL, "ERROR %d: %s\nUnable to resolve hostname or IP address (%s).\n",
                    h_errno, hstrerror(h_errno), utsrv_addr);
            exit(2);
        }
    } else {
        bcopy(*hostinfo->h_addr_list, &dst_addr.sin_addr, sizeof(dst_addr.sin_addr));
    }
    log_msg(DBG, LL_TRACE, "UT server IP address: %s\n",
            inet_ntop(AF_INET, &dst_addr.sin_addr, ip_str, sizeof(ip_str)));

    // TODO: print sensor definition table for reference
    
    // populate unchangeable packet fields
    bzero(&packet, sizeof(packet));
    packet.status = PKT_STAT_INIT;
    get_mac_address(robot_id, packet.robot_id, sizeof(packet.robot_id));
    log_msg(DBG, LL_TRACE, "Robot id (hex): ");
    log_hex(DBG, LL_TRACE, packet.robot_id, sizeof(packet.robot_id));

    // header for logging of sensor data
    log_msg(DAT, LL_INFO, "Timestamp\tSensor\t[ADC]\tValue\t  GPS\tLatitude\t  Longitude\n");

    // main loop: process sensor samples, GPS position and send update messages
    for(;;) {
        
        usleep(SLEEP_USEC);

        // send a packet for each defined sensor
        for(i = 0; i < ADC_CHANNELS; i++) {

            // verify if this ADC channel is defined and usable (i.e. has a sensor attached)
            if(!sensor_defs[i].enabled || 
                sensor_defs[i].sensor_id == NULL ||
                sensor_defs[i].adc_handler == NULL)
                continue;
            
            // get Unix system timestamp (in seconds)
            packet.timestamp = time(NULL);
            
            // get sensor id string from definition table
            bcopy(sensor_defs[i].sensor_id, packet.sensor_id, sizeof(packet.sensor_id));
            
            // read ADC channel
            log_msg(DBG, LL_TRACE, "Reading sensor on ADC channel %d...\n", i);
            if(read_sensor(adc_fd, i, &adc_val) == -1) {
                log_msg(ERR, LL_WARN, "Failed to read sensor on ADC channel %d.\n", i);
                continue;
            }
            
            // compute real sensor value from ADC's discrete output
            packet.sensor_val.f = (*sensor_defs[i].adc_handler)(adc_val);
            log_msg(DBG, LL_TRACE, "Converted ADC#%d reading %u to real value %.2f.\n",
                    i, adc_val, packet.sensor_val.f);
            
            // execute the processing handler, if one is defined
            // (TODO: this feature can be used to trigger an action if value out-of-range)
            if(sensor_defs[i].value_handler != NULL) {
                if(sensor_proc = (*sensor_defs[i].value_handler)(packet.sensor_val.f))
                    log_msg(DBG, LL_DEBUG, "Sensor value out-of-range (channel=%d, value=%f).\n",
                            i, packet.sensor_val.f);
            } else {
                sensor_proc = 0;
            }
            
            // only do further processing if sensor's readings are to be sent to the UT server
            if(sensor_defs[i].transmit) {
                
                // poll GPS daemon to retrieve geographical position
                log_msg(DBG, LL_TRACE, "Polling GPSD server...\n");
                if(gps_poll(gpsdata) == -1) {
                    log_msg(ERR, LL_WARN, "Failed to query GPS...\n");
                    continue;
                }
                
                // determine whether GPS has usable fix or not -- see gpsd(8)
                if(gpsdata->fix.mode >= MODE_2D)
                    SET_BIT(packet.status, PKT_STAT_GPS_FIX);
                else
                    CLR_BIT(packet.status, PKT_STAT_GPS_FIX);
                
                packet.latitude.f = gpsdata->fix.latitude;
                packet.longitude.f = gpsdata->fix.longitude;
                
                // log data for current sensor reading
                log_msg(DAT, LL_INFO, "%u\t%s\t0x%04x\t%.2f\t  %s\t%.6f\t  %.6f\n",
                        packet.timestamp,
                        sensor_defs[i].sensor_id, adc_val, packet.sensor_val.f,
                        (gpsdata->fix.mode >= MODE_2D) ? "yes" : "no",
                        packet.latitude.f, packet.longitude.f);
                
                // send sensor readings to server
                log_msg(DBG, LL_TRACE, "Sending data packet to UT server...\n");
                if(send_packet(sock_fd, &packet, &dst_addr) == -1) {
                    log_msg(ERR, LL_WARN, "Failed to send packet...\n");
                    continue;
                }
            } else {
                // log data for current local sensor reading (data not sent)
                log_msg(DAT, LL_INFO, "%u\t%s\t0x%04x\t%.2f\t  N/A\tN/A\t  N/A\n",
                        packet.timestamp,
                        sensor_defs[i].sensor_id, adc_val, packet.sensor_val.f);
            }
            
            
        } // sensor ADC channels loop

    } // main loop

    return 0;
}
